/***
 * Warthog Robotics
 * University of Sao Paulo (USP) at Sao Carlos
 * http://www.warthog.sc.usp.br/
 *
 * This file is part of WREye project.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ***/

#include "eyeconfiggui.hh"
#include <WREye/GUI/eyeviewgui.hh>
#include <WREye/entity/eyeclient/eyeclient.hh>
#include <WREye/entity/eyeupdater/eyeupdater.hh>
#include <WREye/utils/fieldtype.hh>

using namespace EyeGUI;

EyeConfigGUI::EyeConfigGUI(EyeClient *eyeClient, EyeUpdater *eyeUpdater, EyeViewGUI *eyeView, QWidget *parent) : QWidget(parent) {
    _eyeClient = eyeClient;
    _eyeUpdater = eyeUpdater;
    _eyeView = eyeView;

    // Sets the object name
    this->setObjectName(this->metaObject()->className());

    // Setups the GUI
    setupGUI();

    // Load values to interface
    loadValuesToGUI();

    // Connects the signals
    connectSignals();

    // Set update timer
    _updateTimer.setTimerType(Qt::PreciseTimer);
    _updateTimer.setInterval(200); // 5 Hz
    _updateTimer.start();
    QObject::connect(&_updateTimer, SIGNAL(timeout()), this, SLOT(updateGUI()));
}

EyeConfigGUI::~EyeConfigGUI() {

}

void EyeConfigGUI::setupGUI() {
    // Config layout
    _tabLayout = new QVBoxLayout();
    _tabLayout->setAlignment(Qt::AlignTop);
    this->setLayout(_tabLayout);

    /// INFO GROUP BOX
    // Create
    _groupInfo = new QGroupBox("Vision system information");
    _groupInfo->setFixedHeight(140);

    // Layout
    _groupInfo_layout = new QFormLayout();
    _groupInfo->setLayout(_groupInfo_layout);

    // Create and add rows
    _groupInfo_fieldType = new QLabel("SSL 2015");
    _groupInfo_layout->addRow("EyeView field type:", _groupInfo_fieldType);

    _groupInfo_portNumber = new QLabel("10006");
    _groupInfo_layout->addRow("Vision port number:", _groupInfo_portNumber);

    _groupInfo_numCams = new QLabel("0 camera(s)");
    _groupInfo_layout->addRow("Number of cameras:", _groupInfo_numCams);

    _groupInfo_fpsDetection = new QLabel("0 FPS");
    _groupInfo_layout->addRow("Detection data reception:", _groupInfo_fpsDetection);

    _groupInfo_fpsGeometry = new QLabel("0 FPS");
    _groupInfo_layout->addRow("Geometry data reception:", _groupInfo_fpsGeometry);

    /// FIELD LIMITS
    // Create
    _groupFLimits = new QGroupBox("Field side limit");
    _groupFLimits->setFixedHeight(100);

    // Layout
    _groupFLimits_layout = new QHBoxLayout();
    _groupFLimits->setLayout(_groupFLimits_layout);
    _groupFLimits_layout->setAlignment(Qt::AlignCenter);

    // Create and add buttons
    _groupFLimits_btnQ1 = new QToolButton();
    _groupFLimits_btnQ1->setText("Quadrant 1");
    _groupFLimits_btnQ1->setFixedSize(120, 60);
    _groupFLimits_btnQ1->setCheckable(true);
    _groupFLimits_btnQ1->setIcon(QIcon(":/fieldQ1.png"));
    _groupFLimits_btnQ1->setIconSize(QSize(38, 32));
    _groupFLimits_btnQ1->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    _groupFLimits_layout->addWidget(_groupFLimits_btnQ1);

    _groupFLimits_btnQ2 = new QToolButton();
    _groupFLimits_btnQ2->setText("Quadrant 2");
    _groupFLimits_btnQ2->setFixedSize(120, 60);
    _groupFLimits_btnQ2->setCheckable(true);
    _groupFLimits_btnQ2->setIcon(QIcon(":/fieldQ2.png"));
    _groupFLimits_btnQ2->setIconSize(QSize(38, 32));
    _groupFLimits_btnQ2->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    _groupFLimits_layout->addWidget(_groupFLimits_btnQ2);

    _groupFLimits_btnQ3 = new QToolButton();
    _groupFLimits_btnQ3->setText("Quadrant 3");
    _groupFLimits_btnQ3->setFixedSize(120, 60);
    _groupFLimits_btnQ3->setCheckable(true);
    _groupFLimits_btnQ3->setIcon(QIcon(":/fieldQ3.png"));
    _groupFLimits_btnQ3->setIconSize(QSize(38, 32));
    _groupFLimits_btnQ3->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    _groupFLimits_layout->addWidget(_groupFLimits_btnQ3);

    _groupFLimits_btnQ4 = new QToolButton();
    _groupFLimits_btnQ4->setText("Quadrant 4");
    _groupFLimits_btnQ4->setFixedSize(120, 60);
    _groupFLimits_btnQ4->setCheckable(true);
    _groupFLimits_btnQ4->setIcon(QIcon(":/fieldQ4.png"));
    _groupFLimits_btnQ4->setIconSize(QSize(38, 32));
    _groupFLimits_btnQ4->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    _groupFLimits_layout->addWidget(_groupFLimits_btnQ4);

    _defBtnStylesheet = _groupFLimits_btnQ1->styleSheet();

    /// CONFIG
    // Create
    _groupConfig = new QGroupBox("Configuration");
    _groupConfig->setFixedHeight(100);

    // Layout
    _groupConfig_layout = new QFormLayout();
    _groupConfig->setLayout(_groupConfig_layout);

    // Create and add rows
    _groupConfig_portNumber = new QSpinBox();
    _groupConfig_portNumber->setMinimum(10000);
    _groupConfig_portNumber->setMaximum(10030);
    _groupConfig_portNumber->setValue(10006);
    _groupConfig_layout->addRow("Vision port number:", _groupConfig_portNumber);

    _groupConfig_fieldType = new QComboBox();
	_groupConfig_fieldType->addItem("Vision data");
    _groupConfig_fieldType->addItem("SSL 2012");
    _groupConfig_fieldType->addItem("SSL 2015");
    _groupConfig_fieldType->addItem("SSL 2019");
    _groupConfig_fieldType->addItem("VSS 2008");
	_groupConfig_fieldType->setCurrentIndex(0);
    _groupConfig_layout->addRow("Field type:", _groupConfig_fieldType);


    // Add group boxes
    _tabLayout->addWidget(_groupInfo);
    _tabLayout->addWidget(_groupFLimits);
    _tabLayout->addWidget(_groupConfig);
}

void EyeConfigGUI::connectSignals() {
    QObject::connect(_groupFLimits_btnQ1, SIGNAL(clicked(bool)), this, SLOT(fieldSideLimitBtnQ1Clicked(bool)));
    QObject::connect(_groupFLimits_btnQ2, SIGNAL(clicked(bool)), this, SLOT(fieldSideLimitBtnQ2Clicked(bool)));
    QObject::connect(_groupFLimits_btnQ3, SIGNAL(clicked(bool)), this, SLOT(fieldSideLimitBtnQ3Clicked(bool)));
    QObject::connect(_groupFLimits_btnQ4, SIGNAL(clicked(bool)), this, SLOT(fieldSideLimitBtnQ4Clicked(bool)));
    QObject::connect(_groupConfig_portNumber, SIGNAL(editingFinished()), this, SLOT(portNumberEdited()));
    QObject::connect(_groupConfig_fieldType, SIGNAL(currentTextChanged(QString)), this, SLOT(fieldTypeChanged()));
}

void EyeConfigGUI::loadValuesToGUI() {
    // Port number
    _groupInfo_portNumber->setText(QString::number(_eyeClient->visionPort()));
    _groupConfig_portNumber->setValue(_eyeClient->visionPort());

    // Field type
    _groupInfo_fieldType->setText(FieldTypes::toString(_eyeView->fieldType()));
    _groupConfig_fieldType->setCurrentText(FieldTypes::toString(_eyeView->fieldType()));

    // Field side
    FieldAreas::FieldArea fieldLimit = _eyeUpdater->fieldLimit();
    setBtnChecked(_groupFLimits_btnQ1, FieldAreas::hasArea(fieldLimit, FieldAreas::Q1));
    setBtnChecked(_groupFLimits_btnQ2, FieldAreas::hasArea(fieldLimit, FieldAreas::Q2));
    setBtnChecked(_groupFLimits_btnQ3, FieldAreas::hasArea(fieldLimit, FieldAreas::Q3));
    setBtnChecked(_groupFLimits_btnQ4, FieldAreas::hasArea(fieldLimit, FieldAreas::Q4));
}

void EyeConfigGUI::updateGUI() {
    // Num cameras
    _groupInfo_numCams->setText(QString::number(_eyeClient->numCameras()));

    // Detection FPS
    _groupInfo_fpsDetection->setText(QString::number(_eyeClient->fpsDetection(), 'g', 2)+" FPS");

    // Geometry FPS
    _groupInfo_fpsGeometry->setText(QString::number(_eyeClient->fpsGeometry(), 'g', 2)+" FPS");
}

void EyeConfigGUI::fieldSideLimitBtnQ1Clicked(bool checked) {
    fieldSideLimitBtnClicked(checked, FieldAreas::Q1);
}

void EyeConfigGUI::fieldSideLimitBtnQ2Clicked(bool checked) {
    fieldSideLimitBtnClicked(checked, FieldAreas::Q2);
}

void EyeConfigGUI::fieldSideLimitBtnQ3Clicked(bool checked) {
    fieldSideLimitBtnClicked(checked, FieldAreas::Q3);
}

void EyeConfigGUI::fieldSideLimitBtnQ4Clicked(bool checked) {
    fieldSideLimitBtnClicked(checked, FieldAreas::Q4);
}

void EyeConfigGUI::fieldSideLimitBtnClicked(bool checked, FieldAreas::FieldArea fieldLimit) {
    // Create new area based on click
    FieldAreas::FieldArea currentLimit = _eyeUpdater->fieldLimit();
    FieldAreas::FieldArea newLimit;
    if(checked)
        newLimit = FieldAreas::addArea(currentLimit, fieldLimit);
    else
        newLimit = FieldAreas::delArea(currentLimit, fieldLimit);

    // EyeUpdater change
    _eyeUpdater->setFieldLimit(newLimit);

    // UI change
    loadValuesToGUI();
}

void EyeConfigGUI::setBtnChecked(QToolButton *btn, bool checked) {
    // Set checked
    btn->setChecked(checked);

    // Change border
    if(checked)
        btn->setStyleSheet(btn->styleSheet().append("QToolButton { background-color: #CCC; }"));
    else
        btn->setStyleSheet(_defBtnStylesheet);
}

void EyeConfigGUI::portNumberEdited() {
    int newPortNumber = _groupConfig_portNumber->value();
    if(newPortNumber==_eyeClient->visionPort())
        return;

    // EyeClient change
    _eyeClient->setVisionPort(newPortNumber);

    // UI change
    loadValuesToGUI();
}

void EyeConfigGUI::fieldTypeChanged() {
    QString newFieldTypeStr = _groupConfig_fieldType->currentText();
    FieldTypes::FieldType newFieldType = FieldTypes::fromString(newFieldTypeStr);
    if(newFieldType==_eyeView->fieldType())
        return;

    // EyeView change
    _eyeView->setFieldType(newFieldType);

    // UI change
    loadValuesToGUI();
}
