//========================================================================
//  This software is free: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 3,
//  as published by the Free Software Foundation.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  Version 3 in the file COPYING that came with this distribution.
//  If not, see <http://www.gnu.org/licenses/>.
//========================================================================
/*!
\file    soccerview.cpp
\brief   C++ Implementation: GLSoccerView
\author  Joydeep Biswas (C) 2011
\author  Modified by Warthog Robotics - Division of Computer Vision
*/
//========================================================================

#pragma GCC diagnostic ignored "-Wunused-parameter"

#include "soccerview.h"
#include "util/field_default_constants.h"
#include <WREye/object/wrobjects.hh>

GLSoccerView::FieldDimensions::FieldDimensions() :
  field_length(FieldConstantsRoboCup2015::kFieldLength),
  field_width(FieldConstantsRoboCup2015::kFieldWidth),
  boundary_width(FieldConstantsRoboCup2015::kBoundaryWidth) {
  for (size_t i = 0; i < FieldConstantsRoboCup2015::kNumFieldLines; ++i) {
    lines.push_back(new FieldLine(FieldConstantsRoboCup2015::kFieldLines[i]));
  }
  for (size_t i = 0; i < FieldConstantsRoboCup2015::kNumFieldArcs; ++i) {
    arcs.push_back(new FieldCircularArc(FieldConstantsRoboCup2015::kFieldArcs[i]));
  }
}

GLSoccerView::GLSoccerView(FieldTypes::FieldType fieldType, QWidget* parent) :
    QGLWidget(QGLFormat(QGL::DoubleBuffer | QGL::DepthBuffer | QGL::SampleBuffers),parent) {
  viewScale = (fieldDim.field_length + fieldDim.boundary_width) / sizeHint().width();
  viewScale = max(viewScale, (fieldDim.field_width + fieldDim.boundary_width) / sizeHint().height());

  viewXOffset = viewYOffset = 0.0;
  setAutoFillBackground(false); //Do not let painter auto fill the widget's background: we'll do it manually through openGl
  connect(this, SIGNAL(postRedraw()), this, SLOT(redraw()));
  blueRobotShape = GL_INVALID_VALUE;
  yellowRobotShape = GL_INVALID_VALUE;
  greyRobotShape = GL_INVALID_VALUE;
  blueCircleRobotShape = GL_INVALID_VALUE;
  yellowCircleRobotShape = GL_INVALID_VALUE;
  greyCircleRobotShape = GL_INVALID_VALUE;
  QFont RobotIDFont = this->font();
  RobotIDFont.setWeight(QFont::Bold);
  RobotIDFont.setPointSize(80);
  glText = GLText(RobotIDFont);
  tLastRedraw = 0;

  this->setWindowTitle("[WREye] EyeView");
  this->setWindowIcon(QIcon(":/fieldAll.png"));

  _pathDrawingTimer.start();

  ballsVelocityVec.setAll(0);

  _fieldType = fieldType;
  cleanupFieldLines();
  updateFieldDimensions();
  resetView();
}

GLSoccerView::~GLSoccerView() {
    cleanupFieldLines();
}

void GLSoccerView::updateFieldDimensions() {
    if(_fieldType==FieldTypes::VISION)
		return;

	graphicsMutex.lock();
	if(_fieldType==FieldTypes::VSS2008) {
        fieldDim.field_length = FieldConstantsVSS2008::kFieldLength;
        fieldDim.field_width = FieldConstantsVSS2008::kFieldWidth;
        fieldDim.boundary_width = FieldConstantsVSS2008::kBoundaryWidth;
        fieldDim.lines.clear();
        for(size_t i=0; i<FieldConstantsVSS2008::kNumFieldLines; i++)
          fieldDim.lines.push_back(new FieldLine(FieldConstantsVSS2008::kFieldLines[i]));
        fieldDim.arcs.clear();
        for(size_t i=0; i<FieldConstantsVSS2008::kNumFieldArcs; i++)
          fieldDim.arcs.push_back(new FieldCircularArc(FieldConstantsVSS2008::kFieldArcs[i]));
        fieldDim.triangles.clear();
        for(size_t i=0; i<FieldConstantsVSS2008::kNumFieldTriangles; i++)
          fieldDim.triangles.push_back(new FieldTriangle(FieldConstantsVSS2008::kFieldTriangles[i]));
	} else if(_fieldType==FieldTypes::SSL2012) {
        fieldDim.field_length = FieldConstantsRoboCup2012::kFieldLength;
        fieldDim.field_width = FieldConstantsRoboCup2012::kFieldWidth;
        fieldDim.boundary_width = FieldConstantsRoboCup2012::kBoundaryWidth;
        fieldDim.lines.clear();
        for(size_t i=0; i<FieldConstantsRoboCup2012::kNumFieldLines; i++)
          fieldDim.lines.push_back(new FieldLine(FieldConstantsRoboCup2012::kFieldLines[i]));
        fieldDim.arcs.clear();
        for(size_t i=0; i<FieldConstantsRoboCup2012::kNumFieldArcs; i++)
          fieldDim.arcs.push_back(new FieldCircularArc(FieldConstantsRoboCup2012::kFieldArcs[i]));
        fieldDim.triangles.clear();
	}  else if(_fieldType==FieldTypes::SSL2015) {
        fieldDim.field_length = FieldConstantsRoboCup2015::kFieldLength;
        fieldDim.field_width = FieldConstantsRoboCup2015::kFieldWidth;
        fieldDim.boundary_width = FieldConstantsRoboCup2015::kBoundaryWidth;
        fieldDim.lines.clear();
        for(size_t i=0; i<FieldConstantsRoboCup2015::kNumFieldLines; i++)
          fieldDim.lines.push_back(new FieldLine(FieldConstantsRoboCup2015::kFieldLines[i]));
        fieldDim.arcs.clear();
        for(size_t i=0; i<FieldConstantsRoboCup2015::kNumFieldArcs; i++)
          fieldDim.arcs.push_back(new FieldCircularArc(FieldConstantsRoboCup2015::kFieldArcs[i]));
        fieldDim.triangles.clear();
    } else if(_fieldType==FieldTypes::SSL2019) {
        fieldDim.field_length = FieldConstantsRoboCup2019::kFieldLength;
        fieldDim.field_width = FieldConstantsRoboCup2019::kFieldWidth;
        fieldDim.boundary_width = FieldConstantsRoboCup2019::kBoundaryWidth;
        fieldDim.lines.clear();
        for(size_t i=0; i<FieldConstantsRoboCup2019::kNumFieldLines; i++)
          fieldDim.lines.push_back(new FieldLine(FieldConstantsRoboCup2019::kFieldLines[i]));
        fieldDim.arcs.clear();
        for(size_t i=0; i<FieldConstantsRoboCup2019::kNumFieldArcs; i++)
          fieldDim.arcs.push_back(new FieldCircularArc(FieldConstantsRoboCup2019::kFieldArcs[i]));
        fieldDim.triangles.clear();
    }
	graphicsMutex.unlock();
}

void GLSoccerView::cleanupFieldLines() {
    graphicsMutex.lock();

    for (size_t i = 0; i < fieldDim.lines.size(); ++i)
      delete fieldDim.lines[i];
    fieldDim.lines.clear();

    for (size_t i = 0; i < fieldDim.arcs.size(); ++i)
      delete fieldDim.arcs[i];
    fieldDim.arcs.clear();

    for (size_t i = 0; i < fieldDim.triangles.size(); ++i)
      delete fieldDim.triangles[i];
    fieldDim.triangles.clear();

    graphicsMutex.unlock();
}

void GLSoccerView::redraw()
{
  if(GetTimeSec()-tLastRedraw<MinRedrawInterval)
    return;
  graphicsMutex.lock();
  update();
  graphicsMutex.unlock();
  tLastRedraw = GetTimeSec();
}


void GLSoccerView::mousePressEvent(QMouseEvent* event)
{
  leftButton = event->buttons().testFlag(Qt::LeftButton);
  midButton = event->buttons().testFlag(Qt::MidButton);
  rightButton = event->buttons().testFlag(Qt::RightButton);
//  bool shiftKey = event->modifiers().testFlag(Qt::ShiftModifier);
//  bool ctrlKey = event->modifiers().testFlag(Qt::ControlModifier);

  if(leftButton)
    setCursor(Qt::ClosedHandCursor);
  if(midButton)
    setCursor(Qt::SizeVerCursor);
  if(leftButton || midButton){
    // Start Pan / Zoom
    mouseStartX = event->x();
    mouseStartY = event->y();
    postRedraw();
  }
}

void GLSoccerView::mouseReleaseEvent(QMouseEvent* event)
{
//  bool shiftKey = event->modifiers().testFlag(Qt::ShiftModifier);
//  bool ctrlKey = event->modifiers().testFlag(Qt::ControlModifier);
  setCursor(Qt::ArrowCursor);
}

void GLSoccerView::mouseMoveEvent(QMouseEvent* event)
{
  static const bool debug = false;
  bool leftButton = event->buttons().testFlag(Qt::LeftButton);
  bool midButton = event->buttons().testFlag(Qt::MidButton);
  bool rightButton = event->buttons().testFlag(Qt::RightButton);

  if(debug) printf("MouseMove Event, Left:%d Mid:%d Right:%d\n", leftButton?1:0, midButton?1:0, rightButton?1:0);

  if(leftButton){
    //Pan
    viewXOffset -= viewScale*double(event->x() - mouseStartX);
    viewYOffset += viewScale*double(event->y() - mouseStartY);
    mouseStartX = event->x();
    mouseStartY = event->y();
    recomputeProjection();
    postRedraw();
  }else if(midButton){
    //Zoom
    double zoomRatio = double(event->y() - mouseStartY)/500.0;
//    double oldScale = viewScale;
    viewScale = viewScale*(1.0+zoomRatio);
    recomputeProjection();
    mouseStartX = event->x();
    mouseStartY = event->y();
    postRedraw();
  }
}

void GLSoccerView::wheelEvent(QWheelEvent* event)
{
  static const bool debug = false;
  double zoomRatio = -double(event->delta())/1000.0;
//  double oldScale = viewScale;
  viewScale = viewScale*(1.0+zoomRatio);
  recomputeProjection();
  if(debug) printf("Zoom: %5.3f\n",viewScale);
  postRedraw();
}

void GLSoccerView::keyPressEvent(QKeyEvent* event)
{
  static const bool debug = false;
  if(debug) printf("KeyPress: 0x%08X\n",event->key());
  if(event->key() == Qt::Key_Space)
    resetView();
  if(event->key() == Qt::Key_Escape)
    close();
}

void GLSoccerView::resetView()
{
  viewScale = (fieldDim.field_length + fieldDim.boundary_width) / width();
  viewScale = max(viewScale, (fieldDim.field_width + fieldDim.boundary_width) / height());

  viewXOffset = viewYOffset = 0.0;
  recomputeProjection();
  postRedraw();
}

void GLSoccerView::resizeEvent(QResizeEvent* event)
{
    QGLWidget::resizeEvent(event);
	//Check difference that calling redraw before resetView makes and if it's really necessary
	redraw();

	resetView();
}

void GLSoccerView::recomputeProjection()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-0.54*viewScale*width()+viewXOffset, 0.54*viewScale*width()+viewXOffset, -0.54*viewScale*height()+viewYOffset, 0.54*viewScale*height()+viewYOffset, minZValue, maxZValue);
  glMatrixMode(GL_MODELVIEW);
}

void GLSoccerView::resizeGL(int width, int height)
{
  glViewport(0, 0, width, height);
  recomputeProjection();
}

void GLSoccerView::initializeGL()
{
  if(blueRobotShape!=GL_INVALID_VALUE)
    glDeleteLists(blueRobotShape, 1);
  blueRobotShape = glGenLists(1);
  if(blueRobotShape==GL_INVALID_VALUE){
    printf("Unable to create display list!\n");
    exit(1);
  }
  glNewList(blueRobotShape, GL_COMPILE);
  drawRobot(teamBlue,true,false);
  glEndList();

  if(yellowRobotShape!=GL_INVALID_VALUE)
    glDeleteLists(yellowRobotShape, 1);
  yellowRobotShape = glGenLists(1);
  if(yellowRobotShape==GL_INVALID_VALUE){
    printf("Unable to create display list!\n");
    exit(1);
  }
  glNewList(yellowRobotShape, GL_COMPILE);
  drawRobot(teamYellow,true,false);
  glEndList();

  if(greyRobotShape!=GL_INVALID_VALUE)
    glDeleteLists(greyRobotShape, 1);
  greyRobotShape = glGenLists(1);
  if(greyRobotShape==GL_INVALID_VALUE){
    printf("Unable to create display list!\n");
    exit(1);
  }
  glNewList(greyRobotShape, GL_COMPILE);
  drawRobot(teamUnknown,true,false);
  glEndList();

  if(blueCircleRobotShape!=GL_INVALID_VALUE)
    glDeleteLists(blueCircleRobotShape, 1);
  blueCircleRobotShape = glGenLists(1);
  if(blueRobotShape==GL_INVALID_VALUE){
    printf("Unable to create display list!\n");
    exit(1);
  }
  glNewList(blueCircleRobotShape, GL_COMPILE);
  drawRobot(teamBlue,false,false);
  glEndList();

  if(yellowCircleRobotShape!=GL_INVALID_VALUE)
    glDeleteLists(yellowCircleRobotShape, 1);
  yellowCircleRobotShape = glGenLists(1);
  if(yellowRobotShape==GL_INVALID_VALUE){
    printf("Unable to create display list!\n");
    exit(1);
  }
  glNewList(yellowCircleRobotShape, GL_COMPILE);
  drawRobot(teamYellow,false,false);
  glEndList();

  if(greyCircleRobotShape!=GL_INVALID_VALUE)
    glDeleteLists(greyCircleRobotShape, 1);
  greyCircleRobotShape = glGenLists(1);
  if(greyRobotShape==GL_INVALID_VALUE){
    printf("Unable to create display list!\n");
    exit(1);
  }
  glNewList(greyCircleRobotShape, GL_COMPILE);
  drawRobot(teamUnknown,false,false);
  glEndList();
}


void GLSoccerView::vectorTextTest()
{
  #define TextTest(loc,angle,size,str,halign,valign) \
  {glText.drawString((loc),angle,size,str,halign,valign); \
  vector2d l1,l2; \
  l1.heading(M_PI/180.0*angle); \
  l1 = loc+l1*size*(glText.getWidth(str)); \
  l2.heading(M_PI/180.0*angle); \
  l2 = loc-l2*size*(glText.getWidth(str)); \
  glBegin(GL_LINES); \
  glVertex3d(l1.x,l1.y,9); \
  glVertex3d(l2.x,l2.y,9); \
  glEnd();}

  glColor3d(1,1,1);
  TextTest(vector2d(1,1)*353.6,45,500,"123agdo0",GLText::LeftAligned,GLText::MedianAligned)
  TextTest(vector2d(fieldDim.field_length*0.5,0),0,500,"123agdo0",GLText::RightAligned,GLText::BottomAligned)
  TextTest(vector2d(0,-fieldDim.field_width*0.5),0,500,"123agdo0",GLText::CenterAligned,GLText::TopAligned)
  TextTest(vector2d(-fieldDim.field_length*0.5,0),0,500,"1\ub023agdo0",GLText::CenterAligned,GLText::MiddleAligned)
}

void GLSoccerView::paintEvent(QPaintEvent* event)
{
  graphicsMutex.lock();
  makeCurrent();
  glClearColor(FIELD_COLOR);
  glShadeModel(GL_SMOOTH);
  glDisable(GL_LIGHTING);
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_MULTISAMPLE);

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  drawFieldLines(fieldDim);
  drawRobots();
  drawRobotsVelocities();
  drawBalls();
  drawBallsVelocities();
  writeBallVelocity();
//  vectorTextTest();
  glPopMatrix();
  swapBuffers();
  graphicsMutex.unlock();
}

void GLSoccerView::drawQuad(vector2d loc1, vector2d loc2, double z)
{
  glBegin(GL_QUADS);
  glVertex3d(loc1.x,loc1.y,z);
  glVertex3d(loc2.x,loc1.y,z);
  glVertex3d(loc2.x,loc2.y,z);
  glVertex3d(loc1.x,loc2.y,z);
  glEnd();
}

void GLSoccerView::drawQuad(vector2d v1, vector2d v2, vector2d v3, vector2d v4, double z)
{
  glBegin(GL_QUADS);
  glVertex3d(v1.x,v1.y,z);
  glVertex3d(v2.x,v2.y,z);
  glVertex3d(v3.x,v3.y,z);
  glVertex3d(v4.x,v4.y,z);
  glEnd();
}

void GLSoccerView::drawArc(vector2d loc, double r1, double r2, double theta1, double theta2, double z, double dTheta)
{
  static const double tesselation = 1.0;
  if(dTheta<0){
    dTheta = tesselation/r2;
  }
  glBegin(GL_QUAD_STRIP);
  for(double theta=theta1; theta<theta2; theta+=dTheta){
    double c1 = cos(theta), s1 = sin(theta);
    glVertex3d(r2*c1+loc.x,r2*s1+loc.y,z);
    glVertex3d(r1*c1+loc.x,r1*s1+loc.y,z);
  }
  double c1 = cos(theta2), s1 = sin(theta2);
  glVertex3d(r2*c1+loc.x,r2*s1+loc.y,z);
  glVertex3d(r1*c1+loc.x,r1*s1+loc.y,z);
  glEnd();
}

void GLSoccerView::drawTriangle(vector2d v1, vector2d v2, vector2d v3, double z) {
    glBegin(GL_TRIANGLES);
    glVertex3d(v1.x, v1.y, z);
    glVertex3d(v2.x, v2.y, z);
    glVertex3d(v3.x, v3.y, z);
    glEnd();
}

void GLSoccerView::drawRobot(int team, bool hasAngle, bool useDisplayLists)
{
  if(useDisplayLists){
    switch ( team ){
      case teamBlue:{
        if(hasAngle)
          glCallList(blueRobotShape);
        else
          glCallList(blueCircleRobotShape);
        break;
      }
      case teamYellow:{
        if(hasAngle)
          glCallList(yellowRobotShape);
        else
          glCallList(yellowCircleRobotShape);
        break;
      }
      default:{
        if(hasAngle)
          glCallList(greyRobotShape);
        else
          glCallList(greyCircleRobotShape);
        break;
      }
    }
    return;
  }

  switch ( team ){
    case teamBlue:{
      glColor3d(0.2549, 0.4941, 1.0);
      break;
    }
    case teamYellow:{
      glColor3d(1.0, 0.9529, 0.2431);
      break;
    }
    default:{
      glColor3d(0.5882,0.5882,0.5882);
      break;
    }
  }
  double theta1 = hasAngle?RAD(40):0.0;
  double theta2 = 2.0*M_PI - theta1;
  if(_fieldType==FieldTypes::VSS2008) {
    drawQuad(-35,35,35,-35, RobotZ);
  } else {
      drawArc(0,0,0,90,theta1, theta2, RobotZ);
      if(hasAngle){
        glBegin(GL_TRIANGLES);
        glVertex3d(0,0,RobotZ);
        glVertex3d(90.0*cos(theta1),90.0*sin(theta1),RobotZ);
        glVertex3d(90.0*cos(theta2),90.0*sin(theta2),RobotZ);
        glEnd();
      }
  }

  switch ( team ){
    case teamBlue:{
      glColor3d(0.0706, 0.2314, 0.6275);
      break;
    }
    case teamYellow:{
      glColor3d(0.8, 0.6157, 0.0);
      break;
    }
    default:{
      glColor3d(0.2745,0.2745,0.2745);
      break;
    }
  }
  if(_fieldType==FieldTypes::VSS2008) {
    drawQuad(-40,40,40,35,RobotZ+0.01);
    drawQuad(-40,-40,40,-35,RobotZ+0.01);
    drawQuad(40,-40,35,40,RobotZ+0.01);
    drawQuad(-40,-40,-35,40,RobotZ+0.01);
  } else {
    drawArc(0,0,80,90,theta1, theta2, RobotZ+0.01);
  }
  if(hasAngle) {
    if(_fieldType==FieldTypes::VSS2008) {
      if(team==teamYellow)
        glColor3d(0.7137, 0.5490, 0.0);
      else
        glColor3d(0.2235, 0.1843, 0.4627);
      drawQuad(50.0*cos(theta1)-10, 60.0*sin(theta1), 52.0*cos(theta2), 60.0*sin(theta2), RobotZ+0.01);
    } else
      drawQuad(90.0*cos(theta1)-10,90.0*sin(theta1), 90.0*cos(theta2),90.0*sin(theta2),RobotZ+0.01);
  }
}

void GLSoccerView::drawRobot(vector2d loc, double theta, double conf, int robotID, int team, bool hasAngle)
{
  glPushMatrix();
  glLoadIdentity();
  glTranslated(loc.x,loc.y,0);
  switch ( team ){
    case teamBlue:{
      glColor3d(0.2549, 0.4941, 1.0);
      break;
    }
    case teamYellow:{
      glColor3d(1.0, 0.9529, 0.2431);
      break;
    }
    default:{
      glColor3d(0.5882,0.5882,0.5882);
      break;
    }
  }
  if(_fieldType!=FieldTypes::VSS2008)
    drawQuad(-90,130,-90.0+180.0*conf,160,RobotZ+0.02);
  else
    drawQuad(-45, 50, -45.0+90.0*conf, 60, RobotZ+0.02);
  glColor3d(0.0,0.0,0.0);
  char buf[1024];
  if(robotID!=unknownRobotID)
    snprintf(buf,1023,"%d",robotID);
  else
    snprintf(buf,1023,"?");
  if(_fieldType==FieldTypes::VSS2008)
    glText.drawString(loc,0,50,buf,GLText::CenterAligned,GLText::MiddleAligned);
  else
    glText.drawString(loc,0,100,buf,GLText::CenterAligned,GLText::MiddleAligned);
  switch ( team ){
    case teamBlue:{
      glColor3d(0.0706, 0.2314, 0.6275);
      break;
    }
    case teamYellow:{
      glColor3d(0.8, 0.6157, 0.0);
      break;
    }
    default:{
      glColor3d(0.2745,0.2745,0.2745);
      break;
    }
  }
  if(_fieldType!=FieldTypes::VSS2008) {
    drawQuad(-96,124,96.0,130,RobotZ+0.02);
    drawQuad(-96,124,-90.0,166,RobotZ+0.02);
    drawQuad(-96,160,96.0,166,RobotZ+0.02);
    drawQuad(90,124,96.0,166,RobotZ+0.02);
  } else {
    drawQuad(-47, 47, 47.0, 50, RobotZ+0.02);
    drawQuad(-47, 47,-45.0, 63, RobotZ+0.02);
    drawQuad(-47, 60, 47.0, 63, RobotZ+0.02);
    drawQuad( 44, 50, 47.0, 60, RobotZ+0.02);
  }

  glRotated(theta,0,0,1.0);
  drawRobot(team, hasAngle, true);
  glPopMatrix();
}

void GLSoccerView::drawRobotPath(vector2d loc, int team)
{
    switch ( team ){
      case teamBlue:{
        glColor3d(0.0706, 0.2314, 0.6275);
        break;
      }
      case teamYellow:{
        glColor3d(0.8, 0.6157, 0.0);
        break;
      }
    }
  drawArc(loc,0,7,-M_PI,M_PI,RobotZ-0.01);
  drawArc(loc,8,11,-M_PI,M_PI,RobotZ-0.01);
}

void GLSoccerView::drawFieldLines(FieldDimensions& dimensions)
{
  glColor4f(FIELD_LINES_COLOR);
  for (size_t i = 0; i < fieldDim.lines.size(); ++i) {
    const FieldLine& line = *fieldDim.lines[i];
    const double half_thickness = 0.5 * line.thickness;
    const vector2d p1(line.p1_x, line.p1_y);
    const vector2d p2(line.p2_x, line.p2_y);
    const vector2d perp = (p2 - p1).norm().perp();
//    const vector2d corner1 = p1 - half_thickness * perp;
//    const vector2d corner2 = p2 + half_thickness * perp;
//    drawQuad(corner1, corner2, FieldZ);
    const vector2d v1 = p1 - half_thickness * perp;
    const vector2d v2 = p1 + half_thickness * perp;
    const vector2d v3 = p2 + half_thickness * perp;
    const vector2d v4 = p2 - half_thickness * perp;
    drawQuad(v1, v2, v3, v4, FieldZ);
  }

  for (size_t i = 0; i < fieldDim.arcs.size(); ++i) {
    const FieldCircularArc& arc = *fieldDim.arcs[i];
    const double half_thickness = 0.5 * arc.thickness;
    const double radius = arc.radius;
    const vector2d center(arc.center_x, arc.center_y);
    const double a1 = arc.a1;
    const double a2 = arc.a2;
    drawArc(center, radius - half_thickness, radius + half_thickness, a1, a2, FieldZ);
  }

  for(size_t i=0; i<fieldDim.triangles.size(); i++) {
      const FieldTriangle& triangle = *fieldDim.triangles[i];
      const vector2d v1(triangle.p1_x, triangle.p1_y);
      const vector2d v2(triangle.p2_x, triangle.p2_y);
      const vector2d v3(triangle.p3_x, triangle.p3_y);
      drawTriangle(v1, v2, v3, FieldZ);
  }
}

void GLSoccerView::drawBall(vector2d loc)
{
  glColor3d(1.0,0.5059,0.0);
  drawArc(loc,0,16,-M_PI,M_PI,BallZ);
  glColor3d(0.8706,0.3490,0.0);
  drawArc(loc,15,21,-M_PI,M_PI,BallZ);
}

void GLSoccerView::drawBallPath(vector2d loc)
{
  glColor3d(1.0,0.5059,0.0);
  drawArc(loc,0,7,-M_PI,M_PI,BallZ-0.01);
  glColor3d(0.7411,0.2941,0.0);
  drawArc(loc,8,11,-M_PI,M_PI,BallZ-0.01);
}

void GLSoccerView::drawVector(vector2d v1, vector2d v2, double z) {
    glColor3d(1.0, 0.0, 0.0);

    const vector2d norm = (v2 - v1).norm();
    const vector2d perp = (v2 - v1).norm().perp();

    // Line
    const double quad_half_thickness = 0.5 * 10;
    const vector2d qv1 = v1 - quad_half_thickness * perp;
    const vector2d qv2 = v1 + quad_half_thickness * perp;
    const vector2d qv3 = v2 + quad_half_thickness * perp;
    const vector2d qv4 = v2 - quad_half_thickness * perp;
    drawQuad(qv1, qv2, qv3, qv4, z);

    // Triangle
    const double triangle_thickness = (_fieldType==FieldTypes::VSS2008? 20 : 50);
    const vector2d tv1 = v2 + 2*triangle_thickness*norm;
    const vector2d tv2 = v2 + triangle_thickness*perp;
    const vector2d tv3 = v2 - triangle_thickness*perp;
    drawTriangle(tv1, tv2, tv3, z);
}

void GLSoccerView::drawBalls()
{
  if(balls.empty())
      return;

  // Ball
  drawBall(balls.first());

  // Path
  for(int i=1; i<balls.size(); i++)
        drawBallPath(balls[i]);
}

void GLSoccerView::drawBallsVelocities() {
    if(balls.empty()) {
        return;
    }

	if(ballsVelocityVec.length()>0.025)
		drawVector(balls.first(), balls.first()+ballsVelocityVec*100, BallZ+0.02);
}

void GLSoccerView::drawRobots()
{
    if(robots.empty())
        return;

    // Robots
    for(int j=0; j<robots.first().size(); j++){
      SoccerViewRobot r = robots.first().at(j);
      drawRobot(r.loc,r.angle,r.conf,r.id,r.team,r.hasAngle);
    }

    // Paths
    for(int i=1; i<robots.size(); i++){
        for(int j=0; j<robots[i].size(); j++){
            SoccerViewRobot r = robots[i][j];
            drawRobotPath(r.loc, r.team);
        }
    }
}

void GLSoccerView::drawRobotsVelocities() {
    for(int i=0; i<robotsVelocities.size(); i++) {
        if(robotsVelocities[i].length()>0.025)
            drawVector(robots.first().at(i).loc, robots.first().at(i).loc+robotsVelocities[i]*100, RobotZ+0.02);
    }
}

void GLSoccerView::updateFieldGeometry(const SSL_GeometryFieldSize& fieldSize) {
    if(_fieldType != FieldTypes::VISION)
		return;

    // Cleanup current lines
    cleanupFieldLines();

	graphicsMutex.lock();

    // Lines
	for (int i = 0; i < fieldSize.field_lines_size(); ++i) {
		const SSL_FieldLineSegment& line = fieldSize.field_lines(i);
		fieldDim.lines.push_back(new FieldLine(
			line.name(), line.p1().x(), line.p1().y(),
			line.p2().x(), line.p2().y(), line.thickness()));
	}

    // Arcs
	for (int i = 0; i < fieldSize.field_arcs_size(); ++i) {
		const SSL_FieldCicularArc& arc = fieldSize.field_arcs(i);
		fieldDim.arcs.push_back(new FieldCircularArc(
			arc.name(), arc.center().x(), arc.center().y(),  arc.radius(),
            arc.a1(), arc.a2(), arc.thickness()));
	}

	graphicsMutex.unlock();
}

void GLSoccerView::updateDetection(QList<Robot*> &objRobots, Ball *objBall) {

    graphicsMutex.lock();    

    // Path drawing
    _pathDrawingTimer.stop();
    const bool drawPath = (_pathDrawingTimer.timesec()>=1.0f/PATHDRAWING_FREQUENCY_HZ);
    if(drawPath)
        _pathDrawingTimer.start();

    // Resize
    if(objBall->position().isUnknown()) {
        balls.clear();
    } else {
        if(balls.size()==0)
            balls.resize(1);
    }
    if(robots.size()==0)
        robots.resize(1);

    if(robots.size()>=PATHDRAWING_FRAMES)
        robots.takeAt(1);
    if(balls.size()>=PATHDRAWING_FRAMES)
        balls.takeAt(1);
    robotsVelocities.clear();
	ballsVelocityVec.set(0,0);

    // Robots
    SoccerViewRobot robot;
    robot.angle = 0;
    vector2d velocity;
    QVector<SoccerViewRobot> newRobotsFrame;

    for(int i=0; i<objRobots.size(); i++) {
        Robot* objRobot = objRobots.at(i);

        // Only robots that exists
        if(objRobot->position().isUnknown())
            continue;

        robot.loc.set(objRobot->position().x()*METER2MM, objRobot->position().y()*METER2MM);
        robot.id = objRobot->robotId();
        robot.hasAngle = (objRobot->orientation().isUnknown()==false);
        if(robot.hasAngle)
            robot.angle = DEG(objRobot->orientation().value());
        robot.team = (objRobot->color()==Colors::YELLOW? teamYellow : teamBlue);
        robot.cameraID = 0;
        robot.conf = objRobot->confidence();
        newRobotsFrame.append(robot);

        velocity.set(objRobot->velocity().x(), objRobot->velocity().y());
        robotsVelocities.append(velocity);
    }
    robots.replace(0, newRobotsFrame);
    if(drawPath)
        robots.append(newRobotsFrame);

    // BALL: only if exists
    if(objBall->position().isUnknown()==false) {
        vector2d newBallFrame;

        newBallFrame.set(objBall->position().x()*METER2MM, objBall->position().y()*METER2MM);
		ballsVelocityVec.set(objBall->velocity().x(), objBall->velocity().y());
		ballVelocity = objBall->velocity().abs();

//        float dist = (balls.last()-newBallFrame).length();
        balls.replace(0, newBallFrame);
        if(drawPath) {
//            if(dist>22)
                balls.append(newBallFrame);
//            else if (balls.size()>=2)
//                balls.takeAt(1);
        }
    }

    graphicsMutex.unlock();
    postRedraw();
}

void GLSoccerView::writeBallVelocity() {
	char buf[1024];
	snprintf(buf, 1023, "Ball velocity: %.3lf", ballVelocity);
	glColor3d(1, 1, 1);
	vector2d loc;
	loc.set(-0.5*fieldDim.field_length, -0.54*fieldDim.field_width);
	renderText(loc.x, loc.y, 0.0, buf, this->font());
}

void GLSoccerView::clearDetectionView() {
    graphicsMutex.lock();

    robots.clear();
    robotsVelocities.clear();
    balls.clear();
    ballVelocity = 0.0;
    update();
    graphicsMutex.unlock();

    //    postRedraw();
}

void GLSoccerView::clearGeometryView() {
    graphicsMutex.lock();

    cleanupFieldLines();

    update();
    graphicsMutex.unlock();
}

void GLSoccerView::setFieldType(FieldTypes::FieldType fieldType) {
    _fieldType = fieldType;

    // Robots
    clearDetectionView();
    initializeGL();

    // Field geometry
    cleanupFieldLines();
    updateFieldDimensions();
    resetView();
}
