/***
 * Warthog Robotics
 * University of Sao Paulo (USP) at Sao Carlos
 * http://www.warthog.sc.usp.br/
 *
 * This file is part of WREye project.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ***/

#ifndef EYEUPDATER_H
#define EYEUPDATER_H

#include <QReadWriteLock>
#include <GEARSystem/gearsystem.hh>

#include <WREye/entity/eyebase.hh>
#include <WREye/entity/entity.hh>
#include <WREye/utils/color.hh>
#include <WREye/utils/fieldarea.hh>
#include <3rdparty/robocup_ssl_client.h>
#include "robottracker.hh"
#include "multiobject/multiobjectfilter.hh"

#define MAX_ROBOTS 12
#define MM2METER (1/1000.0)

class EyeUpdater : public Entity {
public:   
    EyeUpdater(EyeClient *eyeClient, FieldAreas::FieldArea fieldLimit=FieldAreas::ALL, bool enableLossFilter=true, bool enableKalmanFilter=true, bool enableNoiseFilter=true, bool debugDetection=false, bool debugGeometry=false);
    ~EyeUpdater();

    QList<Robot*> getRobotObjects() const;
    Ball* getBallObject() const;
    SSL_GeometryFieldSize getFieldGeometry() const;
    bool hasDetectionUpdates(int timeout) const;
    bool hasGeometryUpdates() const;

    void setFieldLimit(FieldAreas::FieldArea fieldLimit) { _fieldLimit = fieldLimit; }
    FieldAreas::FieldArea fieldLimit() const { return _fieldLimit; }

    QString name();
private:
    void initialization();
    void loop();
    void finalization();

    // Internal
    EyeClient *_eyeClient;
    void processDetectionData(const QList<SSL_DetectionFrame> &detectionFrames);
    void processGeometryData(const SSL_GeometryData &geometryData);
    QList<std::pair<int,SSL_DetectionBall> > parseCamerasBalls(const QList<SSL_DetectionFrame> &detectionFrames) const;
    QHash<int,std::pair<int,SSL_DetectionRobot> > parseCamerasRobots(const QList<SSL_DetectionFrame> &detectionFrames) const;
    void processBalls(const QList<std::pair<int,SSL_DetectionBall> > &balls);
    void processRobots(const QHash<int, std::pair<int,SSL_DetectionRobot> > &robots);

    // Server index info
    static const quint8 _blueTeamIndex = Colors::BLUE;      // 0
    static const quint8 _yellowTeamIndex = Colors::YELLOW;  // 1

    // Robot tracker (for robots without id)
    RobotTracker _robotTracker;

    // Multi objects filtering
    QHash<int,QHash<int,MultiObjectFilter*> > _multiFilterRobots;
    MultiObjectFilter *_multiFilterBall;

    // Objects (final vision objects)
    QHash<int,QHash<int,Robot*> > _objRobots; // color -> id -> robot
    Ball *_objBall;

    // Backbone connection (sensor)
    Sensor *_sensor;

    // Sync
    mutable QSemaphore _newDetectionUpdates;
    mutable QSemaphore _newGeometryUpdates;

    // Internal
    FieldAreas::FieldArea _fieldLimit;
    bool _debugDetection;
    bool _debugGeometry;
};

#endif // EYEUPDATER_H
