/***
 * Warthog Robotics
 * University of Sao Paulo (USP) at Sao Carlos
 * http://www.warthog.sc.usp.br/
 *
 * This file is part of WREye project.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ***/

#include "multiobjectfilter.hh"
#include <assert.h>
#include <WREye/utils/utils.hh>
#include <iostream>

MultiObjectFilter::MultiObjectFilter() {

}

MultiObjectFilter::~MultiObjectFilter() {
    reset();
}

void MultiObjectFilter::reset() {
    QList<MultiObject*> objects(_objects.values());

    // Delete multiobjects
    while(objects.empty()==false)
        delete objects.takeFirst();

    // Clear map
    _objects.clear();
}

void MultiObjectFilter::addObject(int camId, float confidence, const Position &pos, const Angle &ori) {
    _objects.insertMulti(confidence, new MultiObject(pos, ori, camId, confidence));
}

QList<MultiObject*> MultiObjectFilter::getFiltered() {

    // Empty or unique object, return
    if(_objects.size()<=1) {
        _noiseFilterTimer.start();
        if(_objects.empty()==false)
            _lastKnownPos = _objects.first()->position();
        return _objects.values();
    }

    // Unify objects
    QMap<float,MultiObject*> unifiedObjects = getUnifiedObjects();

    // Check if has last known position
    if(_lastKnownPos.isUnknown()) {
        MultiObject *retn = unifiedObjects.last();

        // Return the object with best confidence (unified with diff camera secondary objects)
        _noiseFilterTimer.start();
        return getUnification(retn);

    } else {

        // Check if current best ball (higher confidence) is the last ball known
        float dist = WR::Utils::distance(_lastKnownPos, unifiedObjects.last()->position());
        if(dist <= MOFILTER_MERGE_DISTANCE) {
            _noiseFilterTimer.start();
            return getUnification(unifiedObjects.last());
        } else {
            // Check timer
            _noiseFilterTimer.stop();
            if(_noiseFilterTimer.timemsec() >= MOFILTER_THRESHOLD_TIME) { // accept new ball
                _noiseFilterTimer.start();
                return getUnification(unifiedObjects.last());

            } else { // keep old ball
                float minDist=999;
                MultiObject *nearestObject=NULL;

                // Get ball nearest to old position
                for(int i=0; i<unifiedObjects.size(); i++) {
                    MultiObject *object = unifiedObjects.values().at(i);

                    float dist = WR::Utils::distance(object->position(), _lastKnownPos);
                    if(dist < minDist) {
                        minDist = dist;
                        nearestObject = object;
                    }
                }

                // Return
                return getUnification(nearestObject);

            }
        }

    }

}

QMap<float,MultiObject*> MultiObjectFilter::getUnifiedObjects() {
    QMap<float,MultiObject*> unifiedObjects(_objects);

    // Run on object to object
    QMap<float,MultiObject*>::iterator it = unifiedObjects.begin();

    for(int i=0; i<unifiedObjects.size()-1; i++) {
        MultiObject *object1 = unifiedObjects.values().at(i);

        for(int j=i+1; j<unifiedObjects.size(); j++) {
            MultiObject *object2 = unifiedObjects.values().at(j);

            // Discard same camera objects
            if(object1->cameraId()==object2->cameraId())
                continue;

            // Compare and merge, if necessary (diff cams, nearest enough)
            float dist = WR::Utils::distance(object1->position(), object2->position());
            if(dist <= MOFILTER_MERGE_DISTANCE) {
                unifiedObjects.erase(it+j);
                j--;
            }
        }
    }

    return unifiedObjects;
}

QList<MultiObject*> MultiObjectFilter::getUnification(MultiObject *object) {
    QList<MultiObject*> unification(_objects.values());

    // Run on object to object
    for(int i=0; i<unification.size(); i++) {
        MultiObject *obj = unification.at(i);

        // Keep same
        if(obj==object)
            continue;

        // Discard same camera objects
        if(object->cameraId()==obj->cameraId()) {
            unification.removeAt(i);
            i--;
            continue;
        }

        // Remove objects not nearest enough
        float dist = WR::Utils::distance(object->position(), obj->position());
        if(dist > MOFILTER_MERGE_DISTANCE) {
            unification.removeAt(i);
            i--;
            continue;
        }

    }

    _lastKnownPos = object->position();
    return unification;
}
