/***
 * Warthog Robotics
 * University of Sao Paulo (USP) at Sao Carlos
 * http://www.warthog.sc.usp.br/
 *
 * This file is part of WREye project.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ***/

#include "objectvelocity.hh"

ObjectVelocity::ObjectVelocity() {
    distSum.setPosition(0.0, 0.0, 0.0);
}

Velocity ObjectVelocity::getVelocity(const Position &currPos) {
    Velocity retn(true, 0.0, 0.0);

    // Check curr pos
    if(currPos.isUnknown())
        return retn;

    // Curr time
    timespec clock;
    clock_gettime(CLOCK_REALTIME, &clock);
    double currTimeMICRO = (double)((long)clock.tv_sec*1E6 + clock.tv_nsec*1E-3)*1E-6;

    // Add curr frame
    addNewFrame(currPos, currTimeMICRO);

    // Check if exceeded max frames memory
    if(frames.size()>MAX_FRAMES)
        removeOlderFrame();

    // Calc velocity
    Velocity velocity = calcVelocity();
    if(velocity.abs()<0.001f)
        velocity.setVelocity(0.0, 0.0);

    return velocity;
}

void ObjectVelocity::addNewFrame(const Position &pos, double time) {
    // Save last
    ObjectFrame *last=NULL;
    if(frames.isEmpty()==false)
         last = frames.last();

    // Alloc and push
    ObjectFrame *newFrame = new ObjectFrame(pos, time);
    frames.push_back(newFrame);

    // Sum dist
    float x = newFrame->getPos().x() - (last==NULL? newFrame->getPos().x() : last->getPos().x());
    float y = newFrame->getPos().y() - (last==NULL? newFrame->getPos().y() : last->getPos().y());
    distSum.setPosition(distSum.x()+x, distSum.y()+y, 0.0);
}

void ObjectVelocity::removeOlderFrame() {
    if(frames.isEmpty()==false) {
        // Pop
        ObjectFrame *pop = frames.takeFirst();

        // Save new first
        ObjectFrame *first = frames.first();

        // Remove dist
        float x = first->getPos().x() - pop->getPos().x();
        float y = first->getPos().y() - pop->getPos().y();
        distSum.setPosition(distSum.x()-x, distSum.y()-y, 0.0);

        // Delete removed
        delete pop;
    }
}

Velocity ObjectVelocity::calcVelocity() {
    // Get first and last points
    const ObjectFrame *first = (frames.isEmpty())? NULL : frames.first();
    const ObjectFrame *last = (frames.isEmpty())? NULL : frames.last();

    // Calc time between
    double time = (last==NULL? 0 : last->getTime()) - (first==NULL? 0 : first->getTime());

    // Return velocity
    Velocity retn(true, (time==0? 0 : distSum.x()/time), (time==0? 0 : distSum.y()/time));
    return retn;
}
